#!/usr/bin/python3 
# -*-coding: UTF-8 -*-
from array import array
from statistics import stdev, mean, median, mode, StatisticsError
from datetime import datetime, timedelta
import os

"""
Constants that are used to exclude data that result from mouse not completely on the platform, more than one mouse on the platform,
or the load cell not being stable

"""
kMIN_WEIGHT = 15 # weights below this value are excluded
kMAX_WEIGHT = 35 # weights above this value are excluded
kCONST_REPS = 3 # look at this many previous points and this many following points to decide if weight is stable
kCONST_CV = 0.1 # threshold for variance (std/mean) to accept weight as stable, calculated from (n -kCONST_REPS to n + kCONST_REPS)
kBIN_SIZE = 0.203125 # the bin sized used to quantitize the data so a mode (most common value) can be selected. If kBIN_SIZE is a binary fraction, e.g., 13/64, printed values will be easier to read
kTIME_CHUNK = 86400 # the time period, in seconds, over which weight are calculated. Output data will have a weight for each time chunk

"""
LLC_get_weights outputs files for each mouse with data collated from a starting date to an ending date.
folder path is full path to directory where LLC files are stores
outpath_path is the full path to the directory where the result files will be written.
save_mode can be 1 for floating point binary format, 2 for text format, or 3 for both types of output files
"""
def LLC_get_weights (folder_path, start_year, start_month, start_day, end_year, end_month, end_day, output_path, save_mode):
    if not folder_path.endswith ('/'):
        folder_path.append ('/')
    if not output_path.endswith ('/'):
        output_path.append ('/')
    # make datetime objects for start and end time
    start_time = datetime (start_year, start_month, start_day, 0,0,0)
    end_time = datetime (end_year, end_month, end_day, 23,59,59)
    # make a list of file names in the chosen directory
    files=[]
    for file in os.listdir(folder_path):
        if file.startswith ('LCC_'):
            file_time = datetime (int (file.split ('_')[1]), int (file.split ('_')[2]),int (file.split ('_')[3]), 0, 0, 0)
            if file_time >= start_time and file_time <= end_time:
                files.append (file)
    if len (files) == 0: 
        print ('Could not find any load cell data files in the directory', folder_path)
        raise FileNotFoundError
    # sort the file names, which will arrange them by date
    files.sort()
    # start with empty output dictionary. key = mice, value = a dictionary of key =int time, and value = foat array of weights
    sorted_data = {}
    for file in files:
        print (file)
        # time is saved in UNIX time
        tBase = datetime (int (file.split ('_')[1]), int (file.split ('_')[2]),int (file.split ('_')[3]), 0, 0, 0).timestamp()
        # add data for each file
        LLC_get_file_weights (folder_path + file, sorted_data, tBase)
    # iterate through dictionary of mice, making output files for each mouse, containing time, number of entries, and weight, in float or text formats
    date_string = '_' + str (start_year) + '_' + str (start_month) + '_' + str (start_day) + '_' + str (end_year) + '_' + str (end_month) + '_' + str(end_day) 
    for mouse in sorted_data:
        chunk_end = start_time.timestamp() + kTIME_CHUNK
        current_weight = (kMAX_WEIGHT + kMIN_WEIGHT)/2
        temp_results_array = array ('f', []) # fill with results for each time chunk to write to file
        temp_data_array = array ('f', []) # for each time chunk, fill with data from that chunk
        num_entries =0 # save number of entries for each time chunk
        # chunk weight data into temp array based on sorted times
        times=sorted ((sorted_data[mouse]).keys())
        num_times = len (times)
        for i in range (0, num_times - 1):
            # if next time is past chunk_end, write data to temporary results array
            if times [i + 1] > chunk_end:
                # Center of time chunk used for time point
                temp_results_array.append (chunk_end - kTIME_CHUNK/2)
                temp_results_array.append (num_entries)
                if len (temp_data_array) < 2:
                    result = float ('nan')
                else:
                    try:
                        result = mode (temp_data_array)
                    except StatisticsError: # if no true mode, get median of values with same number of occurrences
                        result = modes_median (temp_data_array)
                #print (mouse, times[i], result)
                temp_results_array.append (result)
                current_weight = result
                # zero temporary data array and continue to next time chunk with a clean state
                temp_data_array = array ('f', [])
                num_entries =0
                chunk_end += kTIME_CHUNK
            # append data to temp_data_array
            #print ((sorted_data[mouse])[int(times[i])])
            temp_data_array += (sorted_data[mouse])[int(times[i])]
            num_entries +=1
        # last time point
        temp_results_array.append (chunk_end - kTIME_CHUNK/2)
        temp_results_array.append (num_entries)
        if len (temp_data_array) < 2:
            result = float ('nan')
        else:
            try:
                result = mode (temp_data_array)
            except StatisticsError: # if no true mode, get median of values with same number of occurrences
                result = modes_median (temp_data_array)
        temp_results_array.append (result)
        
        # write results array for this mouse to a file
        if (save_mode & 1):
            file_name = output_path + mouse + date_string + '.dat'
            with open (file_name, 'wb') as out_file:
                temp_results_array.tofile (out_file)
        if (save_mode & 2):
            file_name = output_path + mouse + date_string + '.txt'
            with open (file_name, 'w') as out_file:
                out_file.write ('date and time\tentries\tweight\r')
                for line in range (0, len (temp_results_array), 3):
                    output_str= datetime.fromtimestamp (int (temp_results_array [line])).isoformat (' ') + '\t' 
                    output_str += str (int (temp_results_array[line + 1])) + '\t'
                    output_str += '{:.1f}'.format (temp_results_array[line + 2]) + '\n'                                                       
                    out_file.write (output_str)

        
def LLC_get_file_weights (file, sorted_data, tBase):
 
    # start with an empty array of floats, and append the binary data from the file to it
    data = array ('f', [])
    with open (file, 'rb') as binaryFile:
        try:
            data.frombytes (binaryFile.read())
        except ValueError as e:
            print (str (e))
            return
    # iterate through each epoch of data, doing a preliminary triage and adding data to the dictionary for each mouse
    nPnts = len (data)
    iPt =0
    while iPt < nPnts:
        # first point is mouse RFID code, in negative for easier parsing
        id_code= 'm' + str (int(-data[iPt]))
        # second point is time, in seconds since midnight
        iPt +=1
        time = int (tBase + data[iPt])
        if not id_code in sorted_data:
            sorted_data [id_code] = {time : array ('f', [])}
        else:
            sorted_data [id_code] [time] = array ('f', [])
        # subsequent points til next negative number are weight date for this mouse
        iPt += 1
        startPt=iPt
        while iPt < nPnts and data [iPt] > -10:
            iPt += 1
        endPt = iPt - 1
        # copy good data from this series into sorted data      
        for oPt in range (startPt, endPt, 1):        
            if data [oPt] > kMIN_WEIGHT and data [oPt] < kMAX_WEIGHT:
                running = data [max (oPt-kCONST_REPS, startPt) : min (oPt + kCONST_REPS, endPt)]
                if len (running) >= 2 and (stdev (running)/mean (running) < kCONST_CV):
                    sorted_data [id_code] [time].append(round (data [oPt] /kBIN_SIZE) * kBIN_SIZE)


# if there is a tie for values with most occcurrences, use this function to get the mode closest to previous weight
def modes_closest (arr, current_weight):
    values = array ('f', [])
    occurrences = array ('i', [])
    for value in arr:
        if value in values:
            occurrences [values.index (value)] +=1
        else:
            values.append (value)
            occurrences.append (1)
    sorted_tuples= (sorted(zip(occurrences, values), reverse = True))
    try:
        max_ocurrences = sorted_tuples [0] [0]
        i =0
        best_difference = float ('inf') # can only get better from here
        best_mode = 0
        while i < len (sorted_tuples) and sorted_tuples [i] [0] == max_ocurrences:
            difference = abs (sorted_tuples [i] [1] - current_weight)
            if difference < best_difference:
                best_difference = difference
                best_mode = sorted_tuples [i] [1]
            i += 1
    except IndexError as e:
        print (str (e))
        best_mode = float ('nan')
    return best_mode

# if there is a tie for values with most occcurrences, i.e., the mode, use thus this function to get the median of those values
def modes_median (arr):
    values = array ('f', [])
    occurrences = array ('i', [])
    for value in arr:
        if value in values:
            occurrences [values.index (value)] +=1
        else:
            values.append (value)
            occurrences.append (1)
    sorted_tuples= (sorted(zip(occurrences, values), reverse = True))
    max_ocurrences = sorted_tuples [0] [0]
    modes = array ('f', [])
    i =0
    while i < len (sorted_tuples) and sorted_tuples [i] [0] == max_ocurrences:
        modes.append (sorted_tuples [i] [1])
        i +=1
    return median (modes)

"""
Sample usage of the program
"""
kDATA_FOLDER = '/Users/jamie/Documents/FromPi/amw_data/'
kOUTPUT_FOLDER = '/Users/jamie/Documents/amw_results/'
    
if __name__ == '__main__':
    LLC_get_weights (kDATA_FOLDER, 2016, 11, 6, 2016, 12, 31, kOUTPUT_FOLDER, 2)
